// Copyright (C) 2018-2024, Tellusim Technologies Inc. All rights reserved
// https://tellusim.com/

#version 430 core

/*
 */
#if COMPUTE_SHADER || VERTEX_SHADER
	
	#define REFERENCE	1
	
	struct Vertex {
		vec4 position_normal;
		vec4 normal_texcoord;
	};
	
#endif
	
/*
 */
#if COMPUTE_SHADER
	
	layout(local_size_x = 32) in;
	
	layout(row_major, binding = 0) uniform ComputeParameters {
		int num_vertices;
		int width;
		int height;
		float time;
	};
	
	#if REFERENCE
		layout(std430, binding = 1) readonly buffer VertexBuffer { vec4 vertex_data[]; };
		layout(std430, binding = 0, set = 1) buffer VertexBuffers { vec4 vertex_data[]; } vertex_buffers[];
	#else
		layout(std430, binding = 1) readonly buffer VertexBuffer { Vertex vertex_data[]; };
		layout(std430, binding = 0, set = 1) buffer VertexBuffers { Vertex vertex_data[]; } vertex_buffers[];
	#endif
	
	/*
	 */
	void main() {
		
		ivec2 global_id = gl_GlobalInvocationID.xy;
		
		[[branch]] if(global_id.x < num_vertices && global_id.y < width * height) {
			
			// load vertex
			#if REFERENCE
				Vertex &static_vertex = Vertex(vertex_data, global_id.x * 2);
				vec4 position_normal = static_vertex.position_normal;
				vec4 normal_texcoord = static_vertex.normal_texcoord;
			#else
				vec4 position_normal = vertex_data[global_id.x].position_normal;
				vec4 normal_texcoord = vertex_data[global_id.x].normal_texcoord;
			#endif
			
			// transform vertex
			position_normal.x += float(global_id.y % width) - width * 0.5f + 0.5f;
			position_normal.y += float(global_id.y / width) - height * 0.5f + 0.5f;
			position_normal.z += cos(length(position_normal.xy) * 0.3f + time * 2.0f);
			
			// save vertex
			#if REFERENCE
				Vertex &dynamic_vertex = Vertex(vertex_buffers[global_id.y].vertex_data, global_id.x * 2);
				dynamic_vertex.position_normal = position_normal;
				dynamic_vertex.normal_texcoord = normal_texcoord;
			#else
				vertex_buffers[global_id.y].vertex_data[global_id.x].position_normal = position_normal;
				vertex_buffers[global_id.y].vertex_data[global_id.x].normal_texcoord = normal_texcoord;
			#endif
		}
		
		memoryBarrier();
	}
	
#elif VERTEX_SHADER
	
	layout(row_major, binding = 0) uniform CommonParameters {
		mat4 projection;
		mat4 modelview;
		vec4 camera;
	};
	
	#if REFERENCE
		layout(std430, binding = 1) readonly buffer VertexBuffer { vec4 vertex_data[]; };
		layout(std430, binding = 0, set = 4) readonly buffer VertexBuffers { vec4 vertex_data[]; } vertex_buffers[];
	#else
		layout(std430, binding = 1) readonly buffer VertexBuffer { Vertex vertex_data[]; };
		layout(std430, binding = 0, set = 4) readonly buffer VertexBuffers { Vertex vertex_data[]; } vertex_buffers[];
	#endif
	
	layout(location = 0) flat out int s_instance;
	layout(location = 1) out vec3 s_direction;
	layout(location = 2) out vec3 s_normal;
	layout(location = 3) out vec2 s_texcoord;
	
	/*
	 */
	void main() {
		
		#if REFERENCE
			Vertex &static_vertex = Vertex(vertex_data, gl_VertexIndex * 2);
			Vertex &dynamic_vertex = Vertex(vertex_buffers[nonuniformEXT(gl_InstanceIndex)].vertex_data, gl_VertexIndex * 2);
			vec4 vertex_0 = static_vertex.position_normal;
			vec4 vertex_1 = static_vertex.normal_texcoord;
			vec4 vertex_2 = dynamic_vertex.position_normal;
		#else
			vec4 vertex_0 = vertex_data[gl_VertexIndex].position_normal;
			vec4 vertex_1 = vertex_data[gl_VertexIndex].normal_texcoord;
			vec4 vertex_2 = vertex_buffers[nonuniformEXT(gl_InstanceIndex)].vertex_data[gl_VertexIndex].position_normal;
		#endif
		
		vec4 position = vec4(vertex_2.xyz, 1.0f);
		gl_Position = projection * (modelview * position);
		
		s_instance = gl_InstanceIndex;
		s_direction = camera.xyz - position.xyz;
		s_normal = vec3(vertex_0.w, vertex_1.xy);
		s_texcoord = vertex_1.zw;
	}
	
#elif FRAGMENT_SHADER
	
	layout(binding = 0, set = 2) uniform sampler in_sampler;
	layout(binding = 0, set = 1) uniform texture2D in_texture;
	layout(binding = 0, set = 3) uniform texture2D in_textures[];
	
	layout(location = 0) flat in int s_instance;
	layout(location = 1) in vec3 s_direction;
	layout(location = 2) in vec3 s_normal;
	layout(location = 3) in vec2 s_texcoord;
	
	layout(location = 0) out vec4 out_color;
	
	/*
	 */
	void main() {
		
		vec3 direction = normalize(s_direction);
		vec3 normal = normalize(s_normal);
		
		float diffuse = clamp(dot(direction, normal), 0.0f, 1.0f) * 0.75f;
		float specular = pow(clamp(dot(reflect(-direction, normal), direction), 0.0f, 1.0f), 16.0f);
		
		vec4 color = texture(sampler2D(in_texture, in_sampler), s_texcoord);
		#if CLAY_D3D12
			color *= texture(sampler2D(in_textures[nonuniformEXT(s_instance)], in_sampler), s_texcoord);
		#else
			color *= texture(nonuniformEXT(sampler2D(in_textures[s_instance], in_sampler)), s_texcoord);
		#endif
		
		out_color = color * (diffuse + specular);
	}
	
#endif
